import asyncio
import os
import subprocess
import sys
import time

import discord
from v6d1tokens.client import request_token
from v6d2ctx.context import Benchmark, monitor
from v6d2ctx.handle_content import handle_content
from v6d2ctx.lock_for import lock_for
from v6d2ctx.serve import serve

from v6d3music.app import MusicAppFactory, session_db
from v6d3music.commands import register_commands
from v6d3music.config import prefix
from v6d3music.core.cache_url import cache_db
from v6d3music.core.mainasrc import main_for_raw_vc, mainasrcs
from v6d3music.core.mainaudio import volume_db
from v6d3music.core.queueaudio import queue_db
from v6d3music.utils.entries_effects_for_args import effects_db

loop = asyncio.new_event_loop()
asyncio.set_event_loop(loop)
register_commands()
client = discord.Client(
    intents=discord.Intents(
        members=True,
        guilds=True,
        bans=True,
        emojis=True,
        invites=True,
        voice_states=True,
        guild_messages=True,
        reactions=True
    ),
    loop=loop
)

vcs_restored = False


async def restore_vcs():
    global vcs_restored
    vcs: list[tuple[int, int, bool]] = queue_db.get('vcs', [])
    try:
        for vcgid, vccid, vc_is_paused in vcs:
            try:
                guild: discord.Guild = await client.fetch_guild(vcgid)
                async with lock_for(guild, 'not in a guild'):
                    channels = await guild.fetch_channels()
                    channel: discord.VoiceChannel
                    channel, = [ch for ch in channels if ch.id == vccid]
                    vp: discord.VoiceProtocol = await channel.connect()
                    assert isinstance(vp, discord.VoiceClient)
                    vc = vp
                    await main_for_raw_vc(vc, create=True, force_play=True)
                    if vc_is_paused:
                        vc.pause()
            except Exception as e:
                print(f'vc {vcgid} {vccid} {vc_is_paused} failed', e)
            else:
                print(f'vc restored {vcgid} {vccid}')
    finally:
        vcs_restored = True


@client.event
async def on_ready():
    print('ready')
    await client.change_presence(
        activity=discord.Game(
            name='феноменально',
        )
    )
    async with lock_for('vcs_restored', '...'):
        if not vcs_restored:
            await restore_vcs()


@client.event
async def on_message(message: discord.Message) -> None:
    await handle_content(message, message.content, prefix)


async def save_queues():
    for mainasrc in list(mainasrcs.values()):
        await asyncio.sleep(0.01)
        await mainasrc.queue.save()


async def save_vcs():
    if vcs_restored:
        vcs = []
        vc: discord.VoiceClient
        for vc in list(client.voice_clients):
            await asyncio.sleep(0.01)
            if vc.is_playing():
                if vc.guild is not None and vc.channel is not None:
                    vcs.append((vc.guild.id, vc.channel.id, vc.is_paused()))
        queue_db.set_nowait('vcs', vcs)


async def save_commit():
    await queue_db.set('commit', time.time())


async def save_job():
    while True:
        await asyncio.sleep(1)
        with Benchmark('SVQ'):
            await save_queues()
        with Benchmark('SVV'):
            await save_vcs()
        with Benchmark('SVC'):
            await save_commit()


async def start_app():
    await MusicAppFactory.start(client)


async def setup_tasks():
    loop.create_task(save_job())
    loop.create_task(start_app())


async def main():
    async with volume_db, queue_db, cache_db, session_db, effects_db:
        if 'guerilla' in sys.argv:
            from pathlib import Path
            tokenpath = Path('.token.txt')
            if tokenpath.exists():
                token = tokenpath.read_text()
            else:
                token = input('token:')
                tokenpath.write_text(token)
        else:
            token = await request_token('music', 'token')
        await client.login(token)
        loop.create_task(setup_tasks())
        if os.getenv('v6monitor'):
            loop.create_task(monitor())
        if os.getenv('v6tor', None) is None:
            try:
                subprocess.Popen('tor')
            except FileNotFoundError:
                print('no tor')
        await client.connect()


if __name__ == '__main__':
    serve(main(), client, loop)
